package com.github.gkutiel.flip.db;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import com.sun.xml.internal.messaging.saaj.util.ByteInputStream;

public class FlipDB {
	static class Magic implements InvocationHandler {
		private final List<Connection> pool = new LinkedList<>();
		private final String url;

		public Magic(final String url) {
			this.url = url;
		}

		private PreparedStatement buildStmt(final PreparedStatement ps, final Object[] args) {
			if (args == null) return ps;
			for (int i = 0; i < args.length; i++) {
				final int index = i + 1;
				accept(args[i], new ObjectVisitor() {

					@Override
					public void visit(final byte[] bs) throws SQLException {
						ps.setBinaryStream(index, new ByteInputStream(bs, bs.length));
					}

					@Override
					public void visit(final Double d) throws SQLException {
						ps.setDouble(index, d);
					}

					@Override
					public void visit(final Integer i) throws SQLException {
						ps.setInt(index, i);
					}

					@Override
					public void visit(final Long l) throws SQLException {
						ps.setLong(index, l);
					}

					@Override
					public void visit(final String s) throws SQLException {
						ps.setString(index, s);
					}

					@Override
					public void visitObject(final Object obj) {
						throw new UnsupportedOperationException("Unsupported object type: " + obj.getClass());
					}
				});
			}
			return ps;
		}

		private Connection getConnection() throws SQLException {
			synchronized (pool) {
				if (pool.isEmpty()) return DriverManager.getConnection(url);
				return pool.remove(0);
			}
		}

		@Override
		public Object invoke(final Object proxy, final Method m, final Object[] args) {
			Connection con = null;
			try {
				con = getConnection();
				final PreparedStatement ps = buildStmt(con.prepareStatement(m.getAnnotation(Query.class).value()), args);
				LOGGER.info(ps.toString());
				final Class<?> returnType = m.getReturnType();
				if (returnType.equals(void.class)) {
					ps.executeUpdate();
					return null;
				}
				final ResultSet rs = ps.executeQuery();
				return new ResultSetParser(returnType, rs).parse();
			} catch (final Exception e) {
				LOGGER.warning(e.getMessage());
				return null;
			} finally {
				releseConnection(con);
			}
		}

		private void releseConnection(final Connection con) {
			synchronized (pool) {
				pool.add(con);
			}
		}
	}

	interface ObjectVisitor {

		void visit(byte[] obj) throws SQLException;

		void visit(Double d) throws SQLException;

		void visit(Integer i) throws SQLException;

		void visit(Long l) throws SQLException;

		void visit(String s) throws SQLException;

		void visitObject(Object obj);

	}

	private static final Logger LOGGER = Logger.getLogger(FlipDB.class.getCanonicalName());

	static final Map<Class<?>, String> JAVA_TO_SQL = new HashMap<Class<?>, String>() {
		{
			put(int.class, "INTEGER");
			put(Integer.class, "INTEGER");
			put(long.class, "BIGINT");
			put(Long.class, "BIGINT");
			put(double.class, "DOUBLE");
			put(Double.class, "DOUBLE");
			put(String.class, "VARCHAR");
			put(byte[].class, "BLOB");
		}
	};

	private static void accept(final Object obj, final ObjectVisitor visitor) {
		try {
			if (obj instanceof Integer) visitor.visit((Integer) obj);
			else if (obj instanceof Long) visitor.visit((Long) obj);
			else if (obj instanceof Double) visitor.visit((Double) obj);
			else if (obj instanceof String) visitor.visit((String) obj);
			else if (obj instanceof byte[]) visitor.visit((byte[]) obj);
			else visitor.visitObject(obj);
		} catch (final SQLException e) {
			throw new RuntimeException(e);
		}
	}

	private final String url;

	public FlipDB(@SuppressWarnings("unused") final Driver driver, final String url) {
		this.url = url;
	}

	@SafeVarargs
	final public FlipDB createTables(final Class<? extends Table>... tables) {
		try {
			Connection con = null;
			try {
				con = DriverManager.getConnection(url);
				for (final Class<? extends Table> table : tables) {
					con.createStatement().execute(SqlUtils.createTableSql(table));
					SqlUtils.createIndexes(con, table);
				}
			} finally {
				if (con != null) con.close();
			}
		} catch (final SQLException e) {
			throw new RuntimeException(e);
		}
		return this;
	}

	@SuppressWarnings("unchecked")
	public <T> T getImplementation(final Class<T> forInterface) {
		return (T) Proxy.newProxyInstance(ClassLoader.getSystemClassLoader(), new Class<?>[] { forInterface }, new Magic(url));
	}

}
